# coding: utf-8
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.module_loading import import_string


class XDumpCommand(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument(
            'filename',
            action='store',
            help='Path to dump zip file.',
        )
        parser.add_argument(
            '-a', '--alias',
            action='store',
            dest='alias',
            help='Database configuration alias.',
            required=False,
            default='default',
        )
        parser.add_argument(
            '-b', '--backend',
            action='store',
            dest='backend',
            help='Importable string for custom XDump backend.',
            required=False,
            default=None,
        )

    def handle(self, filename, alias, backend, verbosity, **options):
        backend = self.get_xdump_backend(alias, backend, verbosity)
        self._handle(filename, backend, **options)

    def _handle(self, filename, backend, **options):
        raise NotImplementedError

    def get_xdump_backend(self, alias='default', backend=None, verbosity=0):
        configuration = self.get_database_configuration(alias)
        if backend is None:
            if 'BACKEND' in settings.XDUMP:
                backend = settings.XDUMP['BACKEND']
            else:
                backend = {
                    'django.db.backends.postgresql': 'xdump.postgresql.PostgreSQLBackend',
                    'django.db.backends.postgresql_psycopg2': 'xdump.postgresql.PostgreSQLBackend',
                    'django.db.backends.sqlite': 'xdump.sqlite.SQLiteBackend',
                }[configuration['ENGINE']]
        backend_class = import_string(backend)
        return backend_class(
            dbname=configuration['NAME'],
            user=configuration.get('USER'),
            password=configuration.get('PASSWORD'),
            host=configuration.get('HOST'),
            port=configuration.get('PORT'),
            verbosity=verbosity,
        )

    def get_database_configuration(self, alias):
        return settings.DATABASES[alias]

    def get_dump_kwargs(self):
        return {
            'full_tables': settings.XDUMP['FULL_TABLES'],
            'partial_tables': settings.XDUMP['PARTIAL_TABLES'],
        }



import sys


if sys.version_info[0] == 2:
    from StringIO import StringIO
else:
    from io import StringIO  # noqa

try:
    FileNotFoundError = FileNotFoundError
except NameError:
    FileNotFoundError = OSError

try:
    from functools import lru_cache
except ImportError:
    from repoze.lru import lru_cache as cache

    def lru_cache():

        def wrapper(f):
            wrapped = cache(None)(f)
            wrapped.cache_clear = lambda: wrapped._cache.clear()
            return wrapped

        return wrapper



# coding: utf-8
import os
import sqlite3
import subprocess
import sys
from csv import DictReader, DictWriter

from ._compat import FileNotFoundError, StringIO, lru_cache
from .base import BaseBackend


def dict_factory(cursor, row):
    return {description[0]: value for description, value in zip(cursor.description, row)}


def force_string(value):
    if isinstance(value, bytes):
        value = value.decode()
    return value


class SQLiteBackend(BaseBackend):

    def __init__(self, *args, **kwargs):
        if sqlite3.sqlite_version_info < (3, 8, 3):
            raise RuntimeError('Minimum supported SQLite version is 3.8.3. You have {0}'.format(sqlite3.sqlite_version))
        super(SQLiteBackend, self).__init__(*args, **kwargs)

    def connect(self, *args, **kwargs):
        connection = sqlite3.connect(self.dbname)
        connection.row_factory = dict_factory
        return connection

    def run_dump(self, *args, **kwargs):
        process = subprocess.Popen(('sqlite3', ) + args, stdout=subprocess.PIPE)
        return process.communicate()[0]

    def get_tables_for_related_data(self, full_tables, partial_tables):
        for result in self.run("SELECT name AS table_name FROM sqlite_master WHERE type='table'"):
            yield result['table_name']

    def run(self, sql, params=(), using='default'):
        sql = force_string(sql)
        return super(SQLiteBackend, self).run(sql, params, using)

    def run_many(self, sql):
        with self.log_query(sql):
            sql = force_string(sql)
            cursor = self.get_cursor()
            cursor.executescript(sql)

    def begin_immediate(self):
        cursor = self.get_cursor()
        cursor.execute('BEGIN IMMEDIATE')

    @lru_cache()
    def _get_foreign_keys(self, table):
        return self.run('PRAGMA foreign_key_list({})'.format(table))

    def get_foreign_keys(self, table, full_tables=(), recursive=False):
        for foreign_key in self._get_foreign_keys(table):
            if foreign_key['table'] in full_tables:
                continue
            if foreign_key['table'] == table and not recursive:
                continue
            if foreign_key['table'] != table and recursive:
                continue
            yield {
                'foreign_table_name': foreign_key['table'],
                'table_name': table,
                'foreign_column_name': foreign_key['to'],
                'column_name': foreign_key['from'],
            }
        if sys.version_info[:2] < (3, 6):
            # Before 3.6 sqlite3 used to implicitly commit an open transaction in this case.
            self.begin_immediate()

    def dump(self, filename, full_tables=(), partial_tables=None, **kwargs):
        self.input_check(full_tables, partial_tables)
        self.begin_immediate()
        super(SQLiteBackend, self).dump(filename, full_tables=full_tables, partial_tables=partial_tables, **kwargs)

    def dump_schema(self):
        return self.run_dump(self.dbname, '.schema')

    def export_to_csv(self, sql):
        output = StringIO()
        cursor = self.get_cursor()
        with self.log_query(sql):
            cursor.execute(sql)
            data = cursor.fetchall()
        writer = DictWriter(output, fieldnames=[column[0] for column in cursor.description], lineterminator='\n')
        writer.writeheader()
        writer.writerows(data)
        result = output.getvalue().encode()
        output.close()  # StringIO doesn't support context manager protocol on Python 2
        return result

    def drop_database(self, dbname):
        try:
            os.remove(dbname)
        except FileNotFoundError:
            pass

    def create_database(self, dbname, *args, **kwargs):
        with sqlite3.connect(dbname):
            pass

    def run_setup_file(self, sql):
        self.run_many(sql)

    def load_data(self, archive):
        """
        Loads all data from data files inside the archive to the database.
        """
        for name in archive.namelist():
            if name.startswith(self.data_dir):
                fd = archive.open(name)
                filename = os.path.basename(name).split('.')[0]
                self.load_data_file(filename, fd)

    def load_data_file(self, table_name, fd):
        reader = DictReader(fd.read().decode().split('\n'), delimiter=',')
        fields = ','.join(reader.fieldnames)
        placeholders = ('?,' * len(reader.fieldnames))[:-1]
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO {0} ({1}) VALUES ({2})'.format(table_name, fields, placeholders),
            [[line[k] for k in reader.fieldnames] for line in reader]
        )



# coding: utf-8
import os
import zipfile
from contextlib import contextmanager
from time import time

import attr

from ._compat import lru_cache
from .logging import get_logger


@attr.s(cmp=False)
class BaseBackend(object):
    dbname = attr.ib()
    user = attr.ib()
    password = attr.ib()
    host = attr.ib()
    port = attr.ib(convert=str)
    verbosity = attr.ib(convert=int, default=0)
    connections = {'default': {}}
    schema_filename = 'dump/schema.sql'
    initial_setup_files = (schema_filename, )
    data_dir = 'dump/data/'

    @property
    def logger(self):
        if not hasattr(self, '_logger'):
            self._logger = get_logger('XDump', self.verbosity)
        return self._logger

    @contextmanager
    def log_query(self, sql, params=None):
        self.logger.debug('Execute query: %s' % sql)
        self.logger.debug('Parameters: %s' % str(params))
        with self.log_time():
            yield

    @contextmanager
    def log_time(self, message='Execution time: %s'):
        start = time()
        yield
        self.logger.debug(message % (time() - start))

    # Connection

    @lru_cache()
    def get_connection(self, name='default'):
        return self.connect(**self.connections[name])

    @lru_cache()
    def get_cursor(self, name='default'):
        return self.get_connection(name).cursor()

    def connect(self, *args, **kwargs):
        """
        Create a connection to the database.
        """
        raise NotImplementedError

    def get_connection_kwargs(self, **kwargs):
        for option in ('dbname', 'user', 'password', 'host', 'port'):
            kwargs.setdefault(option, getattr(self, option))
        return kwargs

    def cache_clear(self):
        self.get_cursor.cache_clear()
        self.get_connection.cache_clear()

    # Low-level commands executors

    def run_dump(self, *args, **kwargs):
        """
        Runs built-in utility for data / schema dumping.
        """
        raise NotImplementedError

    def run(self, sql, params=None, using='default'):
        with self.log_query(sql, params):
            cursor = self.get_cursor(using)
            cursor.execute(sql, params)
            try:
                return cursor.fetchall()
            except Exception as exc:
                self.handle_run_exception(exc)

    def handle_run_exception(self, exc):
        raise NotImplementedError

    @contextmanager
    def transaction(self):
        """
        Runs block of code inside a transaction.
        """
        self.run('BEGIN')
        yield
        self.run('COMMIT')

    # Dumping the data

    def dump(self, filename, full_tables=(), partial_tables=None, compression=zipfile.ZIP_DEFLATED, dump_schema=True,
             dump_data=True):
        """
        Creates a dump, which could be used to restore the database.
        """
        self.input_check(full_tables, partial_tables)
        with self.log_time('Total execution time: %s'):
            partial_tables = partial_tables or {}
            with zipfile.ZipFile(filename, 'w', compression) as file:
                if dump_schema:
                    self.write_initial_setup(file)
                if dump_data:
                    self.add_related_data(full_tables, partial_tables)
                    self.write_full_tables(file, full_tables)
                    self.write_partial_tables(file, partial_tables)

    def input_check(self, full_tables, partial_tables):
        if full_tables and partial_tables:
            common_tables = set(full_tables) & set(partial_tables)
            if common_tables:
                raise ValueError(
                    '`partial_tables` should not contain tables from `full_tables`. Common tables: {tables}'.format(
                        tables=', '.join(common_tables)
                    )
                )

    def add_related_data(self, full_tables, partial_tables):
        """
        Updates selects for partial tables to grab all objects, that are referenced by full / partial tables.
        """
        tables = self.get_tables_for_related_data(full_tables, partial_tables)
        for table in tables:
            self.update_partial_tables(table, full_tables, partial_tables)

    def get_tables_for_related_data(self, full_tables, partial_tables):
        return tuple(full_tables) + tuple(partial_tables.keys())

    def update_partial_tables(self, table, full_tables, partial_tables):
        self.update_recursive_relations(table, full_tables, partial_tables)
        self.update_non_recursive_relations(table, full_tables, partial_tables)

    def update_recursive_relations(self, table, full_tables, partial_tables):
        for foreign_key in self.get_foreign_keys(table, full_tables, recursive=True):
            if table in partial_tables:
                partial_tables[table] = RECURSIVE_QUERY_TEMPLATE.format(
                    source=partial_tables[table], target=table, **foreign_key
                )

    def update_non_recursive_relations(self, table, full_tables, partial_tables):
        for foreign_key in self.get_foreign_keys(table, full_tables):
            sql = self.get_related_data_sql(foreign_key, full_tables, partial_tables)
            if sql:
                foreign_table = foreign_key['foreign_table_name']
                if foreign_table in partial_tables:
                    partial_tables[foreign_table] += ' UNION ' + sql
                else:
                    partial_tables[foreign_table] = sql
                # Now we select more than before for given table, so we have to do check related data for it.
                self.update_partial_tables(foreign_table, full_tables, partial_tables)

    def get_foreign_keys(self, table, full_tables=(), recursive=False):
        """
        Looks for foreign keys in the given table. Excluding ones, that will be dumped in ``full_tables``.
        """
        raise NotImplementedError

    def get_related_data_sql(self, foreign_key, full_tables, partial_tables):
        """
        Generates SQL to select related data, that is referred from another table.
        """
        table_name = foreign_key['table_name']
        if table_name in full_tables:
            source = foreign_key['table_name']
        elif table_name in partial_tables:
            source = '({}) T'.format(partial_tables[table_name])
        else:
            return
        return '''
            SELECT
                *
            FROM {foreign_table_name}
            WHERE {foreign_column_name} IN (
                SELECT {column_name} FROM {source}
            )'''.format(source=source, **foreign_key)

    def write_initial_setup(self, file):
        self.write_schema(file)

    def write_schema(self, file):
        """
        Writes a DB schema, functions, etc to the archive.
        """
        schema = self.dump_schema()
        file.writestr(self.schema_filename, schema)

    def write_full_tables(self, file, tables):
        """
        Writes a complete tables dump to the archive.
        """
        for table_name in tables:
            self.write_data_file(file, table_name, 'SELECT * FROM {0}'.format(table_name))

    def write_partial_tables(self, file, config):
        for table_name, sql in config.items():
            self.write_data_file(file, table_name, sql)

    def write_data_file(self, file, table_name, sql):
        data = self.export_to_csv(sql)
        file.writestr('{0}{1}.csv'.format(self.data_dir, table_name), data)

    def export_to_csv(self, sql):
        raise NotImplementedError

    # Database re-creation

    def recreate_database(self, owner=None):
        """
        Drops all connections to the database, drops the database and creates it again.
        """
        if owner is None:
            owner = self.user
        self.drop_database(self.dbname)
        self.create_database(self.dbname, owner)
        self.cache_clear()

    def drop_database(self, dbname):
        raise NotImplementedError

    def create_database(self, dbname, *args, **kwargs):
        raise NotImplementedError

    # Loading the dump

    def load(self, filename):
        """
        Loads schema, sequences and data into the database.
        """
        with self.log_time('Total execution time: %s'):
            archive = zipfile.ZipFile(filename)
            self.initial_setup(archive)
            self.load_data(archive)

    def initial_setup(self, archive):
        """
        Loads schema and initial database configuration.
        """
        for filename in self.initial_setup_files:
            sql = archive.read(filename)
            self.run_setup_file(sql)

    def run_setup_file(self, sql):
        return self.run(sql)

    def load_data(self, archive):
        """
        Loads all data from data files inside the archive to the database.
        """
        with self.transaction():
            for name in archive.namelist():
                if name.startswith(self.data_dir):
                    fd = archive.open(name)
                    table_name = os.path.basename(name).split('.')[0]
                    self.load_data_file(table_name, fd)

    def load_data_file(self, table_name, fd):
        """
        Loads a data file into the database.
        """
        raise NotImplementedError


RECURSIVE_QUERY_TEMPLATE = '''
WITH RECURSIVE recursive_cte AS (
  SELECT * FROM ({source}) S
  UNION
  SELECT T.*
  FROM {table_name} T
  INNER JOIN recursive_cte ON (recursive_cte.{column_name} = T.{foreign_column_name})
)
SELECT * FROM recursive_cte
'''



# coding: utf-8
from ..core import XDumpCommand


class Command(XDumpCommand):
    help = 'Loads an SQL dump.'

    def _handle(self, filename, backend, **options):
        backend.recreate_database()
        backend.load(filename)



# coding: utf-8
from ..core import XDumpCommand


class Command(XDumpCommand):
    help = 'Creates an SQL dump with latest data.'

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            '-d', '--dump_data',
            action='store',
            dest='dump_data',
            help='Control if the data should be dumped.',
            required=False,
            default=True,
        )
        parser.add_argument(
            '-s', '--dump_schema',
            action='store',
            dest='dump_schema',
            help='Control if the schema should be dumped.',
            required=False,
            default=True,
        )

    def _handle(self, filename, backend, **options):
        backend.dump(
            filename, dump_data=options['dump_data'], dump_schema=options['dump_schema'], **self.get_dump_kwargs()
        )



from __future__ import absolute_import

import logging
import sys


DEFAULT_LOGGING_LEVEL = logging.CRITICAL
LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


def get_logger(name, verbosity):
    """
    Returns simple console logger.
    """
    logger = logging.getLogger(name)
    logger.setLevel({
        0: DEFAULT_LOGGING_LEVEL,
        1: logging.INFO,
        2: logging.DEBUG
    }.get(min(2, verbosity), DEFAULT_LOGGING_LEVEL))
    logger.handlers = []
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger.addHandler(handler)
    return logger



# coding: utf-8
import os
import subprocess
from io import BytesIO

import psycopg2
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT, ISOLATION_LEVEL_REPEATABLE_READ
from psycopg2.extras import RealDictConnection

from .base import BaseBackend
from .utils import make_options


SEQUENCES_SQL = "SELECT relname FROM pg_class WHERE relkind = 'S'"
# The query below doesn't use `information_schema.table_constraints` and ``, but instead uses its modified versions
# to mitigate permissions insufficiency on that views (they filter data by permissions of the current user)
# Subqueries for constraints other than FOREIGN KEY are removed as well.
BASE_RELATIONS_QUERY = '''
SELECT
  DISTINCT
  TC.constraint_name,
  TC.table_name,
  KCU.column_name,
  CCU.foreign_table_name,
  CCU.foreign_column_name
FROM
  (
    SELECT
      CN.conname AS constraint_name,
      CL.relname AS table_name
    FROM pg_namespace NS,
      pg_constraint CN,
      pg_class CL
    WHERE
      NS.oid = CL.relnamespace AND
      CN.conrelid = CL.oid AND
      CN.contype = 'f' AND
      CL.relkind = 'r' AND
      NOT pg_is_other_temp_schema(NS.oid)
    ) AS TC
    JOIN information_schema.key_column_usage AS KCU
      ON TC.constraint_name = KCU.constraint_name AND
         KCU.table_name = TC.table_name
    JOIN (
      SELECT
        CL.relname AS foreign_table_name,
        AT.attname AS foreign_column_name,
        CN.conname AS constraint_name
      FROM pg_class CL,
        pg_attribute AT,
        pg_constraint CN
      WHERE
        CL.oid = AT.attrelid AND
        CL.oid = CN.confrelid AND
        AT.attnum = ANY (CN.confkey) AND
        NOT AT.attisdropped AND
        CN.contype = 'f' AND
        CL.relkind = 'r'
      ) AS CCU
    ON CCU.constraint_name = TC.constraint_name
'''


class PostgreSQLBackend(BaseBackend):
    sequences_filename = 'dump/sequences.sql'
    initial_setup_files = BaseBackend.initial_setup_files + (sequences_filename, )
    connections = {
        'default': {
            'isolation_level': ISOLATION_LEVEL_REPEATABLE_READ,
        },
        'maintenance': {
            'dbname': 'postgres',
            'isolation_level': ISOLATION_LEVEL_AUTOCOMMIT,
        }
    }

    def connect(self, isolation_level, **kwargs):
        kwargs = self.get_connection_kwargs(**kwargs)
        connection = psycopg2.connect(**kwargs)
        connection.set_isolation_level(isolation_level)
        return connection

    def get_connection_kwargs(self, **kwargs):
        return super(PostgreSQLBackend, self).get_connection_kwargs(connection_factory=RealDictConnection, **kwargs)

    def handle_run_exception(self, exc):
        """
        Suppress exception when there is nothing to fetch.
        """
        if str(exc) != 'no results to fetch':
            raise exc

    @property
    def run_dump_environment(self):
        environ = os.environ.copy()
        if self.password:
            environ['PGPASSWORD'] = self.password
        return environ

    def run_dump(self, *args, **kwargs):
        process = subprocess.Popen(
            (
                'pg_dump',
                '-U', self.user,
                '-h', self.host,
                '-p', self.port,
                '-d', self.dbname,
            ) + args,
            stdout=subprocess.PIPE,
            env=self.run_dump_environment
        )
        return process.communicate()[0]

    def write_initial_setup(self, file):
        super(PostgreSQLBackend, self).write_initial_setup(file)
        self.write_sequences(file)

    def dump_schema(self):
        """
        Produces SQL for the schema of the database.
        """
        return self.run_dump(
            '-s',  # Schema-only
            '-x',  # Do not dump privileges
        )

    def get_sequences(self):
        """
        To be able to modify our loaded dump we need to load exact sequences states.
        """
        return [row['relname'] for row in self.run(SEQUENCES_SQL)]

    def dump_sequences(self):
        sequences = self.get_sequences()
        return self.run_dump(
            '-a',  # Data-only
            *make_options('-t', sequences)
        )

    def write_sequences(self, file):
        sequences = self.dump_sequences()
        file.writestr(self.sequences_filename, sequences)

    def add_related_data(self, full_tables, partial_tables):
        if full_tables:
            query = BASE_RELATIONS_QUERY + ' WHERE NOT(CCU.foreign_table_name = ANY(%(full_tables)s))'
            kwargs = {'full_tables': list(full_tables)}
        else:
            query = BASE_RELATIONS_QUERY
            kwargs = {}
        self._related_data = self.run(query, kwargs)
        super(PostgreSQLBackend, self).add_related_data(full_tables, partial_tables)

    def get_foreign_keys(self, table, full_tables=(), recursive=False):
        # NOTE, `full_tables` is not used, because it is filtered in `BASE_RELATIONS_QUERY`
        for foreign_key in self._related_data:
            if foreign_key['table_name'] == table:
                if foreign_key['foreign_table_name'] == table and not recursive:
                    continue
                if foreign_key['foreign_table_name'] != table and recursive:
                    continue
                yield foreign_key

    def copy_expert(self, sql, file, **kwargs):
        with self.log_query(sql):
            cursor = self.get_cursor()
            return cursor.copy_expert(sql, file, **kwargs)

    def export_to_csv(self, sql):
        """
        Exports the result of the given sql to CSV with a help of COPY statement.
        """
        with BytesIO() as output:
            self.copy_expert('COPY ({0}) TO STDOUT WITH CSV HEADER'.format(sql), output)
            return output.getvalue()

    def get_search_path(self):
        return self.run('show search_path;')[0]['search_path']

    def restore_search_path(self, search_path):
        self.run("SELECT pg_catalog.set_config('search_path', '{0}', false);".format(search_path))

    def initial_setup(self, archive):
        search_path = self.get_search_path()
        super(PostgreSQLBackend, self).initial_setup(archive)
        self.restore_search_path(search_path)

    def recreate_database(self, owner=None):
        self.drop_connections(self.dbname)
        super(PostgreSQLBackend, self).recreate_database(owner)

    def drop_connections(self, dbname):
        self.run('SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = %s', [dbname], 'maintenance')

    def drop_database(self, dbname):
        self.run('DROP DATABASE IF EXISTS {0}'.format(dbname), using='maintenance')

    def create_database(self, dbname, owner):
        self.run('CREATE DATABASE {0} WITH OWNER {1}'.format(dbname, owner), using='maintenance')

    def load_data_file(self, table_name, fd):
        self.copy_expert('COPY {0} FROM STDIN WITH CSV HEADER'.format(table_name), fd)
